
//
// othello.cc
// Roy van Hal en Walter Kosters
// 4 maart 2020; versie 2020.1
// g++ -Wall -O2 -o othello othello.cc
// Aanroep: ./othello [hoogte] [breedte] [seed] [spellen] [printen 0/1]
//

#include <iostream>
#include <cstdlib>
#include <ctime>
using namespace std;

class Basisspeler;

class othello {
  public: // everything public, who cares :(
    static const int MAXSIZE = 50;
    static const int BUSY = 0;
    static const int BLACK = 1;
    static const int WHITE = 2;
    static const int DRAW = 3;
    static const int RESOURCES = 100000;
    char board[MAXSIZE][MAXSIZE]; // the game board
      // board[0][0] is topleft; board[i][j]: row i, column j
      // empty: '.', White: 'W', Black: 'B'
    int height; // actual height of the board
    int width;  // actual width of the board
    int empties; // number of empty positions
    bool playertomove; // true: Black, false: White
    int lasti, lastj; // last move
    Basisspeler* spelers[2]; // the two AI players

    othello ( ); // default constructor
    othello (int & h, int & w); // other constructor with height and width
    void cleanboard ( ); // clean the board
    void initialposition ( ); // setup four initial pieces
    void print ( ); // print the board
    int gamestatus (int & countW, int & countB); 
      // status of the game: BUSY, BLACK, WHITE or DRAW; count stones
    bool finished ( ); // is game finished?
    int numberofmoves ( ); // how many moves are possible?
    int name (char color); // 'W' --> WHITE, 'B' --> BLACK
    int whois ( ); // who is to move?
    void dothemove (int r); // do r-the move, assuming it exists; r >= 1
    void randommove ( ); // do random move, assuming there is at least one
    void firstmove ( ); // do first move, assuming there is at least one
    void printoutcome ( ); // who won?
    bool canmove (int i, int j); // can player move at (i,j)?
    void actualmove (int i, int j); // we can and will move at (i,j)
    int getwinner(int & countW, int & countB); // who won? how many?
    void initplayers ( ); // initialize players
};//othello

class Basisspeler {
protected:
    othello* spel;
public:
    virtual int volgendeZet ( ) = 0;
};//Basisspeler

// default constructor
othello::othello ( ) {
  height = width = 8;
  cleanboard ( );
}//othello::othello

// other constructor with height and width
othello::othello (int & h, int & w) {
  if ( h > MAXSIZE ) 
    h = MAXSIZE;
  if ( h < 4 )
    h = 4;
  if ( w > MAXSIZE )
    w = MAXSIZE;
  if ( w < 4 )
    w = 4;
  height = h;
  width = w;
  cleanboard ( );
}//othello::othello

// clean the board
void othello::cleanboard ( ) {
  int i, j;
  empties = height * width;
  for ( i = 0; i < height; i++ )
    for ( j = 0; j < width; j++ )
      board[i][j] = '.'; 
  lasti = lastj = -1;
  playertomove = true; // Black begins
  initplayers ( );
  initialposition ( );
}//othello::cleanboard

// setup four initial pieces
void othello::initialposition ( ) {
  int i = ( height - 1 ) / 2, j = ( width - 1 ) / 2;
  empties -= 4;
  board[i][j] = 'B';
  board[i][j+1] = 'W';
  board[i+1][j] = 'W';
  board[i+1][j+1] = 'B';
}//othello::initialposition

// print the board
void othello::print ( ) {
  int i, j;
  for ( i = 0; i < height; i++ ) {
    for ( j = 0; j < width; j++ )
      cout << " " << board[i][j];
    cout << endl;
  }//for
  if ( playertomove )
    cout << "Zwart aan zet (";
  else
    cout << "Wit aan zet   (";
  cout << height * width - empties << " stenen)" << endl << endl;
}//othello::print

// 'W' --> WHITE, 'B' --> BLACK
int othello::name (char color) {
  if ( color == 'W' )
    return WHITE;
  return BLACK;
}//othello::name

// who is to move?
int othello::whois ( ) {
  if ( playertomove )
    return BLACK;
  return WHITE;
}//othello::whois

// status of the game: BUSY, BLACK, WHITE or DRAW
int othello::gamestatus (int & countW, int & countB) {
  int i, j;
  countW = 0, countB = 0;
  for ( i = 0; i < height; i++ )
    for ( j = 0; j < width; j++ )
      if ( board[i][j] == 'W' )
        countW++;
      else if ( board[i][j] == 'B' )
        countB++;
  if ( numberofmoves ( ) > 0 )
    return BUSY;
  playertomove = ! playertomove;
  if ( numberofmoves ( ) > 0 )
    return BUSY;
  playertomove = ! playertomove;
  if ( countW > countB )
    return WHITE;
  else if ( countB > countW )
    return BLACK;
  return DRAW;
}//othello::gamestatus

// how many moves are possible?
int othello::numberofmoves ( ) {
  int i, j, counter = 0;
  for ( i = 0; i < height; i++ )
    for ( j = 0; j < width; j++ )
      if ( canmove (i,j) )
        counter++;
  return counter;
}//othello::numberofmoves

// we can and will move at (i,j)
void othello::actualmove (int i, int j) {
  char color, othercolor;
  int i0, j0;
  if ( playertomove ) {
    color = board[i][j] = 'B'; othercolor = 'W';
  }//if
  else {
    color = board[i][j] = 'W'; othercolor = 'B';
  }//else
  lasti = i;
  lastj = j;
  playertomove = ! playertomove;
  empties--;
  
  i0 = i; j0 = j + 1;
  if ( j0 < width && board[i0][j0] == othercolor ) {
    while ( j0 < width && board[i0][j0] == othercolor ) {
      j0++;
    }//while
    if ( j0 < width && board[i0][j0] == color ) {
      j0--;
      while ( board[i0][j0] == othercolor ) {
        board[i0][j0] = color;
        j0--;
      }//while
    }//if
  }//if

  i0 = i; j0 = j - 1; 
  if ( j0 >= 0 && board[i0][j0] == othercolor ) {
    while ( j0 >= 0 && board[i0][j0] == othercolor ) {
      j0--;
    }//while
    if ( j0 >= 0 && board[i0][j0] == color ) {
      j0++;
      while ( board[i0][j0] == othercolor ) {
        board[i0][j0] = color;
        j0++;
      }//while
    }//if
  }//if

  i0 = i + 1; j0 = j; 
  if ( i0 < height && board[i0][j0] == othercolor ) {
    while ( i0 < height && board[i0][j0] == othercolor ) {
      i0++;
    }//while
    if ( i0 < height && board[i0][j0] == color ) {
      i0--;
      while ( board[i0][j0] == othercolor ) {
        board[i0][j0] = color;
        i0--;
      }//while
    }//if
  }//if

  i0 = i - 1; j0 = j; 
  if ( i0 >= 0 && board[i0][j0] == othercolor ) {
    while ( i0 >= 0 && board[i0][j0] == othercolor ) {
      i0--;
    }//while
    if ( i0 >= 0 && board[i0][j0] == color ) {
      i0++;
      while ( board[i0][j0] == othercolor ) {
        board[i0][j0] = color;
        i0++;
      }//while
    }//if
  }//if

  i0 = i + 1; j0 = j + 1; 
  if ( i0 < height && j0 < width && board[i0][j0] == othercolor ) {
    while ( i0 < height && j0 < width && board[i0][j0] == othercolor ) {
      i0++; j0++;
    }//while
    if ( i0 < height && j0 < width && board[i0][j0] == color ) {
      i0--; j0--;
      while ( board[i0][j0] == othercolor ) {
        board[i0][j0] = color;
        i0--; j0--;
      }//while
    }//if
  }//if
  
  i0 = i - 1; j0 = j - 1; 
  if ( i0 >= 0 && j0 >= 0 && board[i0][j0] == othercolor ) {
    while ( i0 >= 0 && j0 >= 0 && board[i0][j0] == othercolor ) {
      i0--; j0--;
    }//while
    if ( i0 >= 0 && j0 >= 0 && board[i0][j0] == color ) {
      i0++; j0++;
      while ( board[i0][j0] == othercolor ) {
        board[i0][j0] = color;
        i0++; j0++;
      }//while
    }//if
  }//if

  i0 = i + 1; j0 = j - 1; 
  if ( i0 < height && j0 >= 0 && board[i0][j0] == othercolor ) {
    while ( i0 < height && j0 >= 0 && board[i0][j0] == othercolor ) {
      i0++; j0--;
    }//while
    if ( i0 < height && j0 >= 0 && board[i0][j0] == color ) {
      i0--; j0++;
      while ( board[i0][j0] == othercolor ) {
        board[i0][j0] = color;
        i0--; j0++;
      }//while
    }//if
  }//if

  i0 = i - 1; j0 = j + 1; 
  if ( i0 >= 0 && j0 < width && board[i0][j0] == othercolor ) {
    while ( i0 >= 0 && j0 < width && board[i0][j0] == othercolor ) {
      i0--; j0++;
    }//while
    if ( i0 >= 0 && j0 < width && board[i0][j0] == color ) {
      i0++; j0--;
      while ( board[i0][j0] == othercolor ) {
        board[i0][j0] = color;
        i0++; j0--;
      }//while
    }//if
  }//if
}//othello::actualmove

// do the r-th move, assuming it exists; r >= 1
void othello::dothemove (int r) {
  int i, j;
  for ( i = 0; i < height; i++ )
    for ( j = 0; j < width; j++ )
      if ( canmove (i,j) ) {
        r--;
	if ( r == 0 ) {
	  actualmove (i,j);
	  return;
	}//if
      }//if
}//othello:dothemove

// do random move, assuming there is at least one
void othello::randommove ( ) {
  if ( numberofmoves ( ) == 0 ) { // pass!
    playertomove = ! playertomove;
  }//if
  dothemove (rand ( ) % numberofmoves ( ) + 1);
}//othello::randommove

// do first move, assuming there is at least one
void othello::firstmove ( ) {
  if ( numberofmoves ( ) == 0 ) { // pass!
    playertomove = ! playertomove;
  }//if
  dothemove (1);
}//othello::firstmove

// who won?
void othello::printoutcome ( ) {
  int countW, countB;
  int gs = gamestatus (countW, countB);
  if ( gs == BLACK )
    cout << "Black wins!" << endl;
  else if ( gs == WHITE )
    cout << "White wins!" << endl;
  else
    cout << "Draw!" << endl;
}//othello::printoutcome

// who won? 0: BLACK; 1: WHITE; 2: DRAW
int othello::getwinner (int & countW, int & countB) {
  int gs = gamestatus (countW,countB);
  if ( gs == BLACK )
    return 0;
  if ( gs == WHITE )
    return 1;
  return 2;
}//othello::getwinner

// can player (i.e., playertomove) move at (i,j)?
bool othello::canmove (int i, int j) {
  char color, othercolor;
  int i0, j0;
  if ( board[i][j] != '.' )
    return false;
  if ( playertomove ) {
    color = 'B'; othercolor = 'W';
  }//if
  else {
    color = 'W'; othercolor = 'B';
  }//else
    
  i0 = i; j0 = j + 1;
  if ( j0 < width && board[i0][j0] == othercolor ) {
    while ( j0 < width && board[i0][j0] == othercolor ) {
      j0++;
    }//while
    if ( j0 < width && board[i0][j0] == color )
      return true;
  }//if

  i0 = i; j0 = j - 1; 
  if ( j0 >= 0 && board[i0][j0] == othercolor ) {
    while ( j0 >= 0 && board[i0][j0] == othercolor ) {
      j0--;
    }//while
    if ( j0 >= 0 && board[i0][j0] == color )
      return true;
  }//if

  i0 = i + 1; j0 = j; 
  if ( i0 < height && board[i0][j0] == othercolor ) {
    while ( i0 < height && board[i0][j0] == othercolor ) {
      i0++;
    }//while
    if ( i0 < height && board[i0][j0] == color )
      return true;
  }//if

  i0 = i - 1; j0 = j; 
  if ( i0 >= 0 && board[i0][j0] == othercolor ) {
    while ( i0 >= 0 && board[i0][j0] == othercolor ) {
      i0--;
    }//while
    if ( i0 >= 0 && board[i0][j0] == color )
      return true;
  }//if

  i0 = i + 1; j0 = j + 1; 
  if ( i0 < height && j0 < width && board[i0][j0] == othercolor ) {
    while ( i0 < height && j0 < width && board[i0][j0] == othercolor ) {
      i0++; j0++;
    }//while
    if ( i0 < height && j0 < width && board[i0][j0] == color )
      return true;
  }//if
  
  i0 = i - 1; j0 = j - 1; 
  if ( i0 >= 0 && j0 >= 0 && board[i0][j0] == othercolor ) {
    while ( i0 >= 0 && j0 >= 0 && board[i0][j0] == othercolor ) {
      i0--; j0--;
    }//while
    if ( i0 >= 0 && j0 >= 0 && board[i0][j0] == color )
      return true;
  }//if

  i0 = i + 1; j0 = j - 1; 
  if ( i0 < height && j0 >= 0 && board[i0][j0] == othercolor ) {
    while ( i0 < height && j0 >= 0 && board[i0][j0] == othercolor ) {
      i0++; j0--;
    }//while
    if ( i0 < height && j0 >= 0 && board[i0][j0] == color )
      return true;
  }//if

  i0 = i - 1; j0 = j + 1; 
  if ( i0 >= 0 && j0 < width && board[i0][j0] == othercolor ) {
    while ( i0 >= 0 && j0 < width && board[i0][j0] == othercolor ) {
      i0--; j0++;
    }//while
    if ( i0 >= 0 && j0 < width && board[i0][j0] == color )
      return true;
  }//if

  return false;
}//othello::canmove

// is game finished?
bool othello::finished ( ) {
  int countW, countB;
  return ( gamestatus (countW,countB) != BUSY );
}//othello::finished

//beginTODO
#include "othellodommespeler.cc"
#include "othelloMCspeler.cc"

// initialize players: should comply with these two include files
void othello::initplayers ( ) {
  spelers[0] = new Dommespeler (this);
  spelers[1] = new MCspeler (this);
}//othello::initplayers
//eindeTODO


int main (int argc, char* argv[ ]) {
  if ( argc != 6 ) {
    cout << "Aanroepen met: " << endl << argv[0] 
         << " [hoogte] [breedte] [seed] [spellen] [printen 0/1]"
	 << endl;
    return 1;
  }//if
  int hoogte = atoi (argv[1]);
  int breedte = atoi (argv[2]);
  int seed = atoi (argv[3]);
  srand (seed);
  int aantalspellen = atoi (argv[4]);
  bool printen = ( argv[5][0] == '1' );
  int countW, countB;
  int result;
  int wins[3] = {0};
  int p;
  othello B(hoogte,breedte);

  cout << "Er wordt/en " << aantalspellen << " spelletje(s) gespeeld." << endl;
  cout << "Bordgrootte is hoogte = " << hoogte 
       << " bij breedte = " << breedte << "." << endl;
  cout << "De random-seed is " << seed << "." << endl << endl;

  for ( p = 0; p < aantalspellen; p++ ) {
    B.cleanboard ( );
    if ( printen )
      B.print ( );
    do {
      if ( B.numberofmoves ( ) == 0 ) { // pass!
        B.playertomove = ! B.playertomove;
	if ( printen )
          B.print ( );
      }//if
      if ( B.playertomove )
        B.dothemove (B.spelers[0]->volgendeZet ( ));
      else
        B.dothemove (B.spelers[1]->volgendeZet ( ));
      if ( printen )
        B.print ( );
    } while ( ! B.finished ( ) );
    result = B.getwinner (countW,countB);
    if ( printen )
      cout << "Uitslag: " << result
           << " (0: Zwart; 1: Wit; 2: remise)" << endl 
	   << "Zwart: " << countB
	   << "; Wit: " << countW << endl << endl;
    wins[result]++;
  }//for

  cout << "Speler 0 wint: " << wins[0] 
       << ", en Speler 1 wint: " << wins[1]
       << ", en " << wins[2] << " remise(s)." << endl;
  return 0;
}//main

